#!/usr/bin/env python3
########################################################################
# This is the Adaptive CRISPRon model
##########################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow as tf
import sys
from collections import OrderedDict 
import numpy as np
from sklearn.model_selection import train_test_split
from scipy import stats
from sklearn.preprocessing import LabelEncoder
import re
from tensorflow.keras import datasets, layers, models, callbacks, Model, optimizers, Input, utils
from tensorflow.keras.layers import Conv1D, Dropout,LSTM, AveragePooling1D, Flatten, Dense, concatenate, SpatialDropout1D


physical_devices = tf.config.list_physical_devices('GPU')
tf.config.experimental.set_memory_growth(physical_devices[0], True)

#%%
OPT = 'adam' #optimizer (eg. adam)
LEARN = 0.00015#learning rage eg.0.001
EPOCHS = 3000 # N. of epochs, eg 100
SEQ_C = '30mer_gRNA'# get seq from this column
VAL_C = 'Indel_freq_HEK293T' # get value to train from this column 
VAL_G = 'CRISPRoff_score' # get value to train from this column 


BATCH_SIZE= 500 # batch test 
SEED = 37# common random seed (0 equals randomly set)
TYPE = 'CG'# Control whether to use the deltaGb feature, if 'CG', then use, if 'C' then not use

print(
'OPT=%s' % OPT,
'LEARN=%f' % LEARN,
'EPOCHS=%i' % EPOCHS,
'SEQ_C=%s' % SEQ_C,
'VAL_C=%s' % VAL_C,
'VAL_G=%s' % VAL_G,

'BATCH_SIZE=%i' % BATCH_SIZE,
'SEED=%i' % SEED,
'TYPE=%s' % TYPE,
)


#length of input seq
eLENGTH=30
#depth of onehot encoding
eDEPTH=4

def read_seq_file(fn, seq_c0, val_c0, val_g0):
    d = OrderedDict()
    with open(fn, 'rt') as f:
        print(fn)
        head = f.readline().rstrip().split('\t')
        if seq_c0.isnumeric():
            seq_c = int(seq_c0) -1
        else:
            if seq_c0 in head:
                seq_c = head.index(seq_c0)
            else:
                raise Exception

        if val_c0.isnumeric():
            val_c = int(val_c0) -1
        else:
            if val_c0 in head:
                val_c = head.index(val_c0)
            else:
                raise Exception

        if val_g0.isnumeric():
            val_g = int(val_g0) -1
        else:
            if val_g0 in head:
                val_g = head.index(val_g0)
            else:
                raise Exception
        f.seek(0)

        for i,l in enumerate(f):
            v = l.rstrip().split('\t')
            s = v[seq_c]
            if not len(s) == eLENGTH:
                print('"%s" is not a string of length %d in line %d' % (s, eLENGTH, i+1))
                continue
            
            if s in d:
                print('"%s" is not unique in line %d', (s, i+1))
                continue

            try:
                e = float(v[val_c])
            except:
                print('no float value for "%s" in line %d', (s, i+1))
                e = 0.0
                continue
            try:
                g = float(v[val_g])
            except:
                print('no float value for "%s" in line %d', (s, i+1))
                g = 0.0
                continue

            d[s] = [g, e]
 
    return d


def onehot(x):
    z = list()
    for y in list(x):
        if y in "Aa":  z.append(0)
        elif y in "Cc": z.append(1)
        elif y in "Gg": z.append(2)
        elif y in "TtUu": z.append(3)
        else:

            raise Exception
    return z

def set_data(DX, s):
    for j,x in enumerate(onehot(s)):
        DX[j][x] = 1
        
from numpy import loadtxt

#import test index from data preprocessing
testind = loadtxt('test.txt', dtype='int')






#Three mode for different model training, user can choose use Cross Validation, use the Internal Dataset as testset, use random split of the combined dataset, or provide their dataset.
def preprocess_seqwithtest(data,rand = True,Cross=False,CrossandInternal = False):  
    dkeys = data.keys()
    DATA_X = np.zeros((len(dkeys),eLENGTH,eDEPTH), dtype=np.float32) # onehot
    DATA_G = np.zeros((len(dkeys)), dtype=np.float32) #deltaGb
    DATA_Y = np.zeros((len(dkeys)), dtype=np.float32) # efficiency
    for l, s in enumerate(dkeys):
        d = data[s]        
        set_data(DATA_X[l], s)
        DATA_G[l] = d[0]
        DATA_Y[l] = d[1]
    all_indices = list(range(23902))
    for i in testind:
        all_indices.remove(i)
    train_ind, val_ind = train_test_split(all_indices, test_size=1/5,random_state =17)

    if rand:
        DATA_Xtr = DATA_X[train_ind]
        DATA_Gtr = DATA_G[train_ind]
        DATA_Xv = DATA_X[val_ind]
        DATA_Gv = DATA_G[val_ind]
        DATA_Xt = DATA_X[testind]
        DATA_Gt = DATA_G[testind]
        DATA_Ytr = DATA_Y[train_ind]
        DATA_Yv = DATA_Y[val_ind]
        DATA_Yt = DATA_Y[testind]
        return ( DATA_Xtr, DATA_Gtr, DATA_Gv,DATA_Gt, DATA_Ytr, DATA_Xt, DATA_Yt, DATA_Xv, DATA_Yv)        
    
    elif CrossandInternal:
        
        traininds = [ loadtxt(f'train{i}.txt', dtype='int') for i in range(1,7)]
        [train_ind1,train_ind2,train_ind3,train_ind4,train_ind5,train_ind6] = traininds
        DATA_Xtr4 = DATA_X[train_ind4]
        DATA_Xtr2 = DATA_X[train_ind2]    
        DATA_Xtr1 = DATA_X[train_ind1]    

        DATA_Xtr5 = DATA_X[train_ind5]
        DATA_Xtr6 = DATA_X[train_ind6]
        DATA_Gtr4 = DATA_G[train_ind4]
        DATA_Gtr2 = DATA_G[train_ind2]    
        DATA_Gtr1 = DATA_G[train_ind1]    

        DATA_Gtr5 = DATA_G[train_ind5]
        DATA_Gtr6 = DATA_G[train_ind6]
        DATA_Ytr4 = DATA_Y[train_ind4]
        DATA_Ytr2 = DATA_Y[train_ind2]    
        DATA_Ytr1 = DATA_Y[train_ind1]    

        DATA_Ytr5 = DATA_Y[train_ind5]
        DATA_Ytr6 = DATA_Y[train_ind6]
        DATA_Xt = DATA_X[train_ind3]
        DATA_Gt = DATA_G[train_ind3]

        DATA_Yt = DATA_Y[train_ind3]
        Dx = [DATA_Xtr4,DATA_Xtr2, DATA_Xtr1, DATA_Xtr5, DATA_Xtr6]
        Dg = [DATA_Gtr4,DATA_Gtr2, DATA_Gtr1, DATA_Gtr5, DATA_Gtr6]
        Dy = [DATA_Ytr4,DATA_Ytr2, DATA_Ytr1, DATA_Ytr5, DATA_Ytr6]
        return  (Dx,Dg,Dy,DATA_Xt,DATA_Gt,DATA_Yt)
    
    elif Cross:
        traininds = [ loadtxt(f'train{i}.txt', dtype='int') for i in range(1,7)]
        [train_ind1,train_ind2,train_ind3,train_ind4,train_ind5,train_ind6] = traininds
        DATA_Xtr1 = DATA_X[train_ind1]
        DATA_Xtr2 = DATA_X[train_ind2]    
        DATA_Xtr3 = DATA_X[train_ind3]    
        DATA_Xtr4 = DATA_X[train_ind4]
        DATA_Xtr5 = DATA_X[train_ind5]
        DATA_Xtr6 = DATA_X[train_ind6]
        DATA_Gtr1 = DATA_G[train_ind1]
        DATA_Gtr2 = DATA_G[train_ind2]    
        DATA_Gtr3 = DATA_G[train_ind3]    
        DATA_Gtr4 = DATA_G[train_ind4]
        DATA_Gtr5 = DATA_G[train_ind5]
        DATA_Gtr6 = DATA_G[train_ind6]
        DATA_Ytr1 = DATA_Y[train_ind1]
        DATA_Ytr2 = DATA_Y[train_ind2]    
        DATA_Ytr3 = DATA_Y[train_ind3]    
        DATA_Ytr4 = DATA_Y[train_ind4]
        DATA_Ytr5 = DATA_Y[train_ind5]
        DATA_Ytr6 = DATA_Y[train_ind6]
        DATA_Xt = DATA_X[testind]
        DATA_Gt = DATA_G[testind]
    
        DATA_Yt = DATA_Y[testind]
        Dx = [DATA_Xtr1,DATA_Xtr2, DATA_Xtr3,DATA_Xtr4, DATA_Xtr5, DATA_Xtr6]
        Dg = [DATA_Gtr1,DATA_Gtr2, DATA_Gtr3, DATA_Gtr4, DATA_Gtr5, DATA_Gtr6]
        Dy = [DATA_Ytr1,DATA_Ytr2, DATA_Ytr3, DATA_Ytr4, DATA_Ytr5, DATA_Ytr6]
        return  (Dx,Dg,Dy,DATA_Xt,DATA_Gt,DATA_Yt )
    
    else:
        traininds = [ loadtxt(f'train{i}.txt', dtype='int') for i in range(1,7)]
        [train_ind1,train_ind2,train_ind3,train_ind4,train_ind5,train_ind6] = traininds
        trainind = [*train_ind1,*train_ind2,*train_ind3,*train_ind4,*train_ind5]
        val_ind = train_ind6
        DATA_Xtr = DATA_X[trainind]
        DATA_Gtr = DATA_G[trainind]
        DATA_Xv = DATA_X[val_ind]
        DATA_Gv = DATA_G[val_ind]
        DATA_Xt = DATA_X[testind]
        DATA_Gt = DATA_G[testind]
        DATA_Ytr = DATA_Y[trainind]
        DATA_Yv = DATA_Y[val_ind]
        DATA_Yt = DATA_Y[testind]
            
        return ( DATA_Xtr, DATA_Gtr, DATA_Gv,DATA_Gt, DATA_Ytr, DATA_Xt, DATA_Yt, DATA_Xv, DATA_Yv)






# check of inputs and setting a few base parameters

if OPT =='adam':
    optimizer = optimizers.Adam(LEARN)
elif OPT == 'rmsprop':
    optimizer = optimizers.RMSprop(LEARN)
else:
    raise Exception

inputs = list()
if TYPE.find('C') > -1:
    input_c = Input(shape=(eLENGTH, eDEPTH,), name="input_onehot")
    inputs.append(input_c)

#delta Gb
if TYPE.find('G') > -1:
    input_g = Input(shape=(1,), name="input_dGB")
    inputs.append(input_g)


for_dense = list()

# Build block recursively but unused for Adap CRISPRon 
# def COVBlock(inp,name,size,num):
#     conv_out = Conv1D(num,size, activation='relu', input_shape=(eLENGTH,eDEPTH,), name=name+"conv_3")(inp)
#     conv_dropout_out = Dropout(0.3, name=name+"drop_3")(conv_out)
#     conv_pool_out = AveragePooling1D(2, padding='SAME', name=name+'pool_3')(conv_dropout_out)
#     conv_flatten_out = Flatten(name=name+"flatten_3")(conv_pool_out)
#     return conv_flatten_out

#first convolution block
if TYPE.find('C') > -1:
    number = [100,70,40]
    size = [3,5,7]
    # for_dense = [COVBlock(input_c, '{}'.format(i), size[i] , number[i]) for i in range(len(size))]
    conv1_out = Conv1D(140, 5, activation='relu', input_shape=(eLENGTH,4,), name="conv_1")(input_c)
    conv1_dropout_out = Dropout(0.3, name="drop_1")(conv1_out)
    conv1_pool_out = AveragePooling1D(2, padding='SAME', name="pool_1")(conv1_dropout_out)
    conv1_pool_out = Conv1D( 60, 3, activation='relu',  name="conv_1_first_descendant")(conv1_pool_out)
    conv1_pool_out = Dropout(0.3, name="drop_1_first_descendant")(conv1_pool_out)
    conv1_pool_out = AveragePooling1D(2, padding='SAME',name="pool_1_first_descendant")(conv1_pool_out)
    conv1_pool_out = Conv1D(60 ,5, activation='relu', name="conv_1_second_descendant")(conv1_pool_out)
    conv1_pool_out = Dropout(0.3, name="drop_1_second_descendant")(conv1_pool_out)
    conv1_pool_out = AveragePooling1D(2, padding='SAME',name="pool_1_second_descendant")(conv1_pool_out)
    conv1_flatten_out = Flatten(name="flatten_1")(conv1_pool_out)
    
    for_dense.append(conv1_flatten_out)

#second convolution block
    conv2_out = Conv1D(60, 6, activation='relu', input_shape=(eLENGTH,4,), name="conv_2")(input_c)
    conv2_dropout_out = Dropout(0.3, name="drop_2")(conv2_out)
    conv2_pool_out = AveragePooling1D(2, padding='SAME', name="pool_2")(conv2_dropout_out)
    conv2_flatten_out = Flatten(name="flatten_2")(conv2_pool_out)
    for_dense.append(conv2_flatten_out)

#third convolution block
    conv3_out = Conv1D(20, 24, activation='relu', input_shape=(eLENGTH,4,), name="conv_3")(input_c)
    conv3_dropout_out = Dropout(0.3, name="drop_3")(conv3_out)
    conv3_pool_out = AveragePooling1D(2, padding='SAME', name="pool_3")(conv3_dropout_out)
    conv3_flatten_out = Flatten(name="flatten_3")(conv3_pool_out)
    for_dense.append(conv3_flatten_out)


    
#concatenation of conv layers and deltaGb layer
if len(for_dense) == 1:
    concat_out = for_dense[0]
else:
    concat_out = concatenate(for_dense)

for_dense1 = list()

#first dense (fully connected) layer
dense0_out = Dense(80, activation='relu', name="dense_0")(concat_out)
dense0_dropout_out = Dropout(0.3, name="drop_d0")(dense0_out)
for_dense1.append(dense0_dropout_out)

print(dense0_dropout_out)
#%%
#Gb input used raw
if TYPE.find('G') > -1:
    for_dense1.append(input_g)


if len(for_dense1) == 1:
    concat1_out = for_dense1[0]
else:
    concat1_out = concatenate(for_dense1)


#first dense (fully connected) layer
dense1_out = Dense(80, activation='relu', name="dense_1")(concat1_out)
dense1_dropout_out = Dropout(0.3, name="drop_d1")(dense1_out)

#second dense (fully connected) layer
dense2_out = Dense(60, activation='relu', name="dense_2")(dense1_dropout_out)
dense2_dropout_out = Dropout(0.3, name="drop_d2")(dense2_out)

#output layer
output = Dense(1, name="output")(dense2_dropout_out)

#model construction

#mc = callbacks.ModelCheckpoint('.model.best', verbose=1, save_best_only=True)
d = read_seq_file('trainingwithoff.tsv', SEQ_C, VAL_C, VAL_G)
(x, g, gv,gt, y, xt, yt,xv, yv) = preprocess_seqwithtest(d,rand=True)

predicts = []
pv = []

es = callbacks.EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=100,restore_best_weights = True)
# logdir="logs/fit/" + datetime.now().strftime("%Y%m%d-%H%M%S")
# tensorboard_callback = keras.callbacks.TensorBoard(log_dir=logdir)
model = Model(inputs=inputs, outputs=[output]) 
model.compile(loss='mse', optimizer=optimizer)
model.save_weights('model.h5')

tinput = list()
vinput = list()
ttinput = list()
if TYPE.find('C') > -1:
    tinput.append(x)
    vinput.append(xv)
    ttinput.append(xt)
if TYPE.find('G') > -1:
    tinput.append(g)
    vinput.append(gv)
    ttinput.append(gt)
tinput = tuple(tinput)
vinput = tuple(vinput)
ttinput = tuple(ttinput)

history = model.fit(tinput, y, validation_data=(vinput, yv), batch_size=BATCH_SIZE, \
        epochs=EPOCHS, use_multiprocessing=True, workers=16, verbose=2, callbacks=[es,tensorboard_callback])
# ,tensorboard_callback
pred = model.predict(ttinput)

sp = stats.spearmanr(pred,yt)
print(sp)

